@@ -15,6 +15,26 @@
  * You should have received a copy of the GNU General Public License
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  *----------------------------------------------------------------------------
+ * Version 1.11
+ *   - Add option to not split lines, if the rest isn't a minimum
+ *     of 3 (default) frames long
+ *   - Images are now autocropped by default
+ *   - Images are now converted to 8bit palette by default
+ *   - Misc bugfixes that should've been fixed before
+ *
+ * Version 1.10b
+ *   - Splitting of images is now disabled by default
+ *
+ * Version 1.10
+ *   - Add option to split events longer than a given number of frames, to
+ *     allow better seeking
+ *   - Change image filename format to XXXXXXXX_X.png
+ *   - Add option to optimize for Presentation Graphics buffer usage, by
+ *     auto-cropping images, and splitting them into two separate ones to
+ *     minimize surface area, when necessary (enabled by default)
+ *   - Use getopts for commandline argument parsing
+ *   - Add Makefile
+ *
  * Version 1.9
  *   - Correct calculation of timecodes. Integer only, now.
  *   - Remove support for drop timecodes. If anybody really needs them, and
@@ -67,6 +87,9 @@
 #include <stdlib.h>
 #include <time.h>
 #include <png.h>
+#include <getopt.h>
+#include "auto_split.h"
+#include "palletize.h"
 
 /* AVIS input code taken from muxers.c from the x264 project (GPLv2 or later).
  * Authors: Laurent Aimar <fenrir@via.ecp.fr>
@@ -185,16 +208,21 @@ int close_file_avis( avis_input_t *handle )
 
 /* Main avs2bdnxml code starts here, too */
 
-void write_png(int file_id, uint8_t *image, int w, int h)
+void write_png(int file_id, uint8_t *image, int w, int h, int graphic, uint32_t *pal, crop_t c)
 {
 	FILE *fh;
 	png_structp png_ptr;
 	png_infop info_ptr;
 	png_bytep *row_pointers;
-	char filename[13];
+	png_colorp palette = NULL;
+	png_bytep trans = NULL;
+	char filename[15];
+	char *col;
+	int step = pal == NULL ? 4 : 1;
+	int colors = 0;
 	int i;
 
-	snprintf(filename, 13, "%08d.png", file_id);
+	snprintf(filename, 15, "%08d_%d.png", file_id, graphic);
 
 	if ((fh = fopen(filename, "wb")) == NULL)
 	{
@@ -232,15 +260,35 @@ void write_png(int file_id, uint8_t *image, int w, int h)
 	png_init_io(png_ptr, fh);
 
 	/* Set file info */
-	png_set_IHDR(png_ptr, info_ptr, w, h, 8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
+	if (pal == NULL)
+		png_set_IHDR(png_ptr, info_ptr, c.w, c.h, 8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
+	else
+	{
+		png_set_IHDR(png_ptr, info_ptr, c.w, c.h, 8, PNG_COLOR_TYPE_PALETTE, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
+		palette = calloc(256, sizeof(png_color));
+		trans = calloc(256, sizeof(png_byte));
+		colors = 1;
+		for (i = 1; i < 256 && pal[i]; i++)
+		{
+			col = (char *)&(pal[i]);
+			palette[i].red = col[0];
+			palette[i].green = col[1];
+			palette[i].blue = col[2];
+			trans[i] = col[3];
+			colors++;
+		}
+		png_set_PLTE(png_ptr, info_ptr, palette, colors);
+		png_set_tRNS(png_ptr, info_ptr, trans, colors, NULL);
+	}
 
 	/* Allocate row pointer memory */
-	row_pointers = calloc(h, sizeof(png_bytep));
+	row_pointers = calloc(c.h, sizeof(png_bytep));
 
 	/* Set row pointers */
-	for (i = 0; i < h; i++)
+	image = image + step * (c.x + w * c.y);
+	for (i = 0; i < c.h; i++)
 	{
-		row_pointers[i] = image + i * w * 4;
+		row_pointers[i] = image + i * w * step;
 	}
 	png_set_rows(png_ptr, info_ptr, row_pointers);
 
@@ -254,6 +302,12 @@ void write_png(int file_id, uint8_t *image, int w, int h)
 	/* Free memory */
 	png_destroy_write_struct(&png_ptr, &info_ptr);
 	free(row_pointers);
+	if (pal != NULL)
+		free(pal);
+	if (palette != NULL)
+		free(palette);
+	if (trans != NULL)
+		free(trans);
 
 	/* Close file handle */
 	fclose(fh);
@@ -365,8 +419,8 @@ void swap_rb_sse2 (stream_info_t *s_info, char *img, char *out)
 	/* Set up masks */
 	asm volatile
 	(
-		"movdqa (%%eax), %%xmm0\n"
-		"movdqa (%%ecx), %%xmm5\n"
+		"movupd (%%eax), %%xmm0\n"
+		"movupd (%%ecx), %%xmm5\n"
 		:
 		: "a" (ag_mask), "c" (rb_mask)
 	);
@@ -502,17 +556,17 @@ int is_empty (stream_info_t *s_info, char *img)
 void zero_transparent (stream_info_t *s_info, char *img)
 {
 	if (detect_sse2())
-		zero_transparent_sse2(s_info, img);
+		return zero_transparent_sse2(s_info, img);
 	else
-		zero_transparent_c(s_info, img);
+		return zero_transparent_c(s_info, img);
 }
 
 void swap_rb (stream_info_t *s_info, char *img, char *out)
 {
 	if (detect_sse2())
-		swap_rb_sse2(s_info, img, out);
+		return swap_rb_sse2(s_info, img, out);
 	else
-		swap_rb_c(s_info, img, out);
+		return swap_rb_c(s_info, img, out);
 }
 
 /* SMPTE non-drop time code */
@@ -532,7 +586,7 @@ void mk_timecode (int frame, int fps, char *buf) /* buf must have length 12 (inc
 
 	if (h > 99)
 	{
-		fprintf(stderr, "Timecodes above 99:59:59:99 not supported: %Lu:%02u:%02u:%02u\n", h, m, s, frames);
+		fprintf(stderr, "Timecodes above 99:59:59:99 not supported: %u:%02u:%02u:%02u\n", h, m, s, frames);
 		exit(1);
 	}
 
@@ -545,21 +599,38 @@ void mk_timecode (int frame, int fps, char *buf) /* buf must have length 12 (inc
 
 void print_usage ()
 {
-	fprintf(stderr, "Usage: avs2bdnxml INPUT TRACKNAME LANGUAGE VIDEOFORMAT FRAMERATE OUTPUT\n\n"
-		"INPUT\t\tInput file (AviSynth script)\n"
-		"TRACKNAME\tName of track, like: Undefined\n"
-		"LANGUAGE\tLanguage code, like: und\n"
-		"VIDEOFORMAT\tEither of: 480i, 480p, 576i, 720p, 1080i, 1080p\n"
-		"FRAMERATE\tEither of: 23.976, 24, 25, 29.97, 50, 59.94\n"
-		/*"         \tFor drop timecodes, use: 23.976d, 29.97d, 59.94d\n"*/
-		"OUTPUT\t\tOutput file in BDN XML format\n\n"
-		"Example: avs2bdnxml input.avs Undefined und 1080p 23.976 output.xml\n");
+	fprintf(stderr,
+		"Usage: avs2bdnxml [options] -o output input\n\n"
+		"Input has to be an AviSynth script with RGBA as output colorspace\n\n"
+		"  -o, --output <string>        Output file in BDN XML format\n"
+		"  -t, --trackname <string>     Name of track, like: Undefined\n"
+		"  -l, --language <string>      Language code, like: und\n"
+		"  -v, --video-format <string>  Either of: 480i, 480p,  576i,\n"
+		"                                          720p, 1080i, 1080p\n"
+		"  -f, --fps <float>            Either of: 23.976, 24, 25, 29.97, 50, 59.94\n"
+		"  -x, --x-offset <integer>     X offset, for use with partial frames.\n"
+		"  -y, --y-offset <integer>     Y offset, for use with partial frames.\n"
+		"  -s, --split-at <integer>     Split events longer than this, in frames.\n"
+		"                               Disabled when 0, which is the default.\n"
+		"  -m, --min-split <integer>    Minimum length of line segment after split.\n"
+		"  -a, --autocrop <integer>     Automatically crop output. [on=1, off=0]\n"
+		"  -p, --palette <integer>      Output 8bit palette PNG. [on=1, off=0]\n"
+		"  -b, --buffer-opt <integer>   Optimize PG buffer size by image\n"
+		"                               splitting. [on=1, off=0]\n\n"
+		"Example:\n"
+		"  avs2bdnxml -t Undefined -l und -v 1080p -f 23.976 -a1 -81 -b0 -m3 \\\n"
+		"    -o output.xml input.avs\n"
+		"  (Input and output are required settings. The rest are set to default.)\n"
+		);
 }
 
 struct event_list_node_s
 {
+	int image_number;
 	int start_frame;
 	int end_frame;
+	int graphics;
+	crop_t c[2];
 	struct event_list_node_s *next;
 };
 
@@ -569,11 +640,15 @@ typedef struct
 	struct event_list_node_s *end;
 } event_list_head_t;
 
-void add_event_xml(event_list_head_t *events, int start, int end)
+void add_event_xml_real (event_list_head_t *events, int image, int start, int end, int graphics, crop_t *crops)
 {
 	struct event_list_node_s *new = calloc(1, sizeof(struct event_list_node_s));
+	new->image_number = image;
 	new->start_frame = start;
 	new->end_frame = end;
+	new->graphics = graphics;
+	new->c[0] = crops[0];
+	new->c[1] = crops[1];
 	if (events->start == NULL)
 	{
 		events->start = new;
@@ -586,6 +661,26 @@ void add_event_xml(event_list_head_t *events, int start, int end)
 	}
 }
 
+void add_event_xml (event_list_head_t *events, int split_at, int min_split, int start, int end, int graphics, crop_t *crops)
+{
+	int image = start;
+	int d = end - start;
+
+	if (!split_at)
+		add_event_xml_real(events, image, start, end, graphics, crops);
+	else
+	{
+		while (d >= split_at + min_split)
+		{
+			d -= split_at;
+			add_event_xml_real(events, image, start, start + split_at, graphics, crops);
+			start += split_at;
+		}
+		if (d)
+			add_event_xml_real(events, image, start, start + d, graphics, crops);
+	}
+}
+
 struct framerate_entry_s
 {
 	char *name;
@@ -608,20 +703,43 @@ int main (int argc, char *argv[])
 	                                        /*, {"59.94d", "59.94", 60000/1001.0, 1}*/
 	                                        , {NULL, NULL, 0, 0}
 	                                        };
-	char *avs_filename, *track_name, *language, *video_format, *frame_rate, *out_filename;
-	char *in_img, *old_img, *tmp, *out_buf;
-	char *intc_buf, *outtc_buf;
-	char *drop_frame;
-	int fps;
+	char *avs_filename = NULL;
+	char *track_name = "Undefined";
+	char *language = "und";
+	char *video_format = "1080p";
+	char *frame_rate = "23.976";
+	char *out_filename = NULL;
+	char *x_offset = "0";
+	char *y_offset = "0";
+	char *buffer_optimize = "0";
+	char *split_after = "0";
+	char *minimum_split = "3";
+	char *palletize_png = "1";
+	char *auto_crop_image = "1";
+	char *in_img = NULL, *old_img = NULL, *tmp = NULL, *out_buf = NULL;
+	char *intc_buf = NULL, *outtc_buf = NULL;
+	char *drop_frame = NULL;
+	crop_t crops[2];
+	pic_t pic;
+	uint32_t *pal = NULL;
+	int have_fps = 0;
+	int n_crop = 1;
+	int split_at = 0;
+	int min_split = 3;
+	int autocrop = 0;
+	int xo, yo;
+	int fps = 25;
 	int frames;
-	int first_frame = -1, start_frame, end_frame;
+	int first_frame = -1, start_frame = -1, end_frame = -1;
 	int num_of_events = 0;
-	int i;
+	int i, c, j;
 	int have_line = 0;
-	int must_zero;
+	int must_zero = 0;
 	int checked_empty;
 	int auto_cut = 0;
+	int pal_png = 1;
 	int progress_step = 1000;
+	int buffer_opt;
 	int bench_start = time(NULL);
 	avis_input_t *avis_hnd;
 	stream_info_t *s_info = malloc(sizeof(stream_info_t));
@@ -630,35 +748,125 @@ int main (int argc, char *argv[])
 	FILE *fh;
 
 	/* Get args */
-	if (argc < 7)
+	if (argc < 2)
 	{
 		print_usage();
 		return 0;
 	}
-	avs_filename = argv[1];
-	track_name = argv[2];
-	language = argv[3];
-	video_format = argv[4];
-	out_filename = argv[6];
+	while (1)
+	{
+		static struct option long_options[] =
+			{ {"output",       required_argument, 0, 'o'}
+			, {"trackname",    required_argument, 0, 't'}
+			, {"language",     required_argument, 0, 'l'}
+			, {"video-format", required_argument, 0, 'v'}
+			, {"fps",          required_argument, 0, 'f'}
+			, {"x-offset",     required_argument, 0, 'x'}
+			, {"y-offset",     required_argument, 0, 'y'}
+			, {"split-at",     required_argument, 0, 's'}
+			, {"min-split",    required_argument, 0, 'm'}
+			, {"autocrop",     required_argument, 0, 'a'}
+			, {"palette",      required_argument, 0, 'p'}
+			, {"buffer-opt",   required_argument, 0, 'b'}
+			, {0, 0, 0, 0}
+			};
+			int option_index = 0;
+
+			c = getopt_long(argc, argv, "o:t:l:v:f:x:y:b:s:m:p:a:", long_options, &option_index);
+			if (c == -1)
+				break;
+			switch (c)
+			{
+				case 'o':
+					out_filename = optarg;
+					break;
+				case 't':
+					track_name = optarg;
+					break;
+				case 'l':
+					language = optarg;
+					break;
+				case 'v':
+					video_format = optarg;
+					break;
+				case 'f':
+					frame_rate = optarg;
+					break;
+				case 'x':
+					x_offset = optarg;
+					break;
+				case 'y':
+					y_offset = optarg;
+					break;
+				case 'p':
+					palletize_png = optarg;
+					break;
+				case 'a':
+					auto_crop_image = optarg;
+					break;
+				case 'b':
+					buffer_optimize = optarg;
+					break;
+				case 's':
+					split_after = optarg;
+					break;
+				case 'm':
+					minimum_split = optarg;
+					break;
+				default:
+					print_usage();
+					return 0;
+					break;
+			}
+	}
+	if (argc - optind == 1)
+		avs_filename = argv[optind];
+	else
+	{
+		fprintf(stderr, "Only a single input file allowed.\n");
+		return 1;
+	}
+
+	/* Both input and output filenames are required */
+	if (avs_filename == NULL)
+	{
+		print_usage();
+		return 0;
+	}
+	if (out_filename == NULL)
+	{
+		print_usage();
+		return 0;
+	}
+
+	/* Set X and Y offsets, and split value */
+	xo = atoi(x_offset);
+	yo = atoi(y_offset);
+	pal_png = atoi(palletize_png);
+	autocrop = atoi(auto_crop_image);
+	split_at = atoi(split_after);
+	min_split = atoi(minimum_split);
+	if (!min_split)
+		min_split = 1;
 
 	/* TODO: Sanity check video_format and frame_rate. */
 
 	/* Get frame rate */
-	frame_rate = NULL;
 	i = 0;
 	while (framerates[i].name != NULL)
 	{
-		if (!strcasecmp(framerates[i].name, argv[5]))
+		if (!strcasecmp(framerates[i].name, frame_rate))
 		{
 			frame_rate = framerates[i].out_name;
 			fps = framerates[i].rate;
 			drop_frame = framerates[i].drop ? "true" : "false";
+			have_fps = 1;
 		}
 		i++;
 	}
-	if (frame_rate == NULL)
+	if (!have_fps)
 	{
-		fprintf(stderr, "Error: Invalid framerate (%s).\n", argv[5]);
+		fprintf(stderr, "Error: Invalid framerate (%s).\n", frame_rate);
 		return 1;
 	}
 
@@ -675,12 +883,24 @@ int main (int argc, char *argv[])
 	old_img = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 2, sizeof(char)); /* see above */
 	out_buf = calloc(s_info->i_width * s_info->i_height * 4 + 16 * 2, sizeof(char));
 
-	/* align buffers */
+	/* Align buffers */
 	in_img  = in_img + (16 - ((int)in_img % 16));
 	old_img = old_img + (16 - ((int)old_img % 16));
 	out_buf = out_buf + (16 - ((int)out_buf % 16));
 
-	/* get frame number */
+	/* Set up buffer (non-)optimization */
+	buffer_opt = atoi(buffer_optimize);
+	pic.b = out_buf;
+	pic.w = s_info->i_width;
+	pic.h = s_info->i_height;
+	pic.s = s_info->i_width;
+	n_crop = 1;
+	crops[0].x = 0;
+	crops[0].y = 0;
+	crops[0].w = pic.w;
+	crops[0].h = pic.h;
+
+	/* Get frame number */
 	frames  = get_frame_total_avis(avis_hnd);
 
 	/* No frames mean nothing to do */
@@ -743,7 +963,7 @@ int main (int argc, char *argv[])
 		/* Not a dup, write end-of-line, if we had a line before */
 		if (have_line)
 		{
-			add_event_xml(events, start_frame, i);
+			add_event_xml(events, split_at, min_split, start_frame, i, n_crop, crops);
 			end_frame = i;
 			have_line = 0;
 		}
@@ -761,7 +981,20 @@ int main (int argc, char *argv[])
 		have_line = 1;
 		start_frame = i;
 		swap_rb(s_info, in_img, out_buf);
-		write_png(start_frame, out_buf, s_info->i_width, s_info->i_height);
+		if (buffer_opt)
+			n_crop = auto_split(pic, crops);
+		else if (autocrop)
+		{
+			crops[0].x = 0;
+			crops[0].y = 0;
+			crops[0].w = pic.w;
+			crops[0].h = pic.h;
+			auto_crop(pic, &(crops[0]));
+		}
+		if (pal_png)
+			pal = palletize(out_buf, s_info->i_width, s_info->i_height);
+		for (j = 0; j < n_crop; j++)
+			write_png(start_frame, out_buf, s_info->i_width, s_info->i_height, j, pal, crops[j]);
 		num_of_events++;
 		if (first_frame == -1)
 			first_frame = i;
@@ -777,7 +1010,7 @@ int main (int argc, char *argv[])
 	/* Add last event, if available */
 	if (have_line)
 	{
-		add_event_xml(events, start_frame, i - 1);
+		add_event_xml(events, split_at, min_split, start_frame, i - 1, n_crop, crops);
 		auto_cut = 1;
 		end_frame = i - 1;
 	}
@@ -835,7 +1068,12 @@ int main (int argc, char *argv[])
 				mk_timecode(node->end_frame + 1, fps, outtc_buf);
 			}
 			
-			fprintf(fh, "<Event Forced=\"False\" InTC=\"%s\" OutTC=\"%s\">\n<Graphic Width=\"%d\" Height=\"%d\" X=\"0\" Y=\"0\">%08d.png</Graphic>\n</Event>\n", intc_buf, outtc_buf, s_info->i_width, s_info->i_height, node->start_frame);
+			fprintf(fh, "<Event Forced=\"False\" InTC=\"%s\" OutTC=\"%s\">\n", intc_buf, outtc_buf);
+			for (i = 0; i < node->graphics; i++)
+			{
+				fprintf(fh, "<Graphic Width=\"%d\" Height=\"%d\" X=\"%d\" Y=\"%d\">%08d_%d.png</Graphic>\n", node->c[i].w, node->c[i].h, xo + node->c[i].x, yo + node->c[i].y, node->image_number, i);
+			}
+			fprintf(fh, "</Event>\n");
 			node = node->next;
 		}
 		while (node != NULL);
@@ -849,7 +1087,9 @@ int main (int argc, char *argv[])
 	close_file_avis(avis_hnd);
 
 	/* Give runtime */
-	/*fprintf(stderr, "Time elapsed: %d\n", time(NULL) - bench_start);*/
+	if (0)
+		fprintf(stderr, "Time elapsed: %lu\n", time(NULL) - bench_start);
 
 	return 0;
 }
+
